/* Dynamic classes
   Bob Gray
   Gisli Hjalmtysson
   Network Mathematics Research Department (AK0112730)
   AT&T
   600 Mountain Avenue
   Murray Hill, New Jersey

   4 June 1996

   dynexcept.h

   This file defines the exceptions associated with class "dynamic".

   Revision history:

   30 Nov 1999: Added a copy constructor and assignment operator to all
                exception classes that contain a message string (needless to 
                say, a terrible oversight on our part).
*/

#ifndef _DYN_EXCEPT_H
#define _DYN_EXCEPT_H

#include "exception.h"
#include "dynversion.h"

class NoDownload: public Exception
{
        char *URL;

    public:

	NoDownload (const char *URL);
	NoDownload (const NoDownload &exception);
	NoDownload& operator= (const NoDownload &exception);
	char *describeException (void);
       ~NoDownload ();
};
 
class ZombieVersion: public Exception
{
    public:

	char *describeException (void);
};

class InvalidHandle: public Exception
{
    public:

	char *describeException (void);
};

class ActiveVersion: public Exception
{
    public:

	char *describeException (void);
};

class NoActiveVersions: public Exception
{
    public:

	char *describeException (void);
}; 

class NoVersionLibrary: public Exception
{
	char *dirName;

    public:

        NoVersionLibrary (const char *dirName);
	NoVersionLibrary (const NoVersionLibrary &exception);
	NoVersionLibrary& operator= (const NoVersionLibrary &exception);
	char *describeException (void);
       ~NoVersionLibrary ();
};

class NoTemporaryLibrary: public Exception
{
	char *dirName;

    public:

        NoTemporaryLibrary (const char *dirName);
	NoTemporaryLibrary (const NoTemporaryLibrary &exception);
	NoTemporaryLibrary& operator= (const NoTemporaryLibrary &exception);
	char *describeException (void);
       ~NoTemporaryLibrary ();
};

class InvalidVersion: public Exception
{
	char *className;
	char *pathName;
	char *dlmessage;	  /* dynamic linker message */

    public:

	InvalidVersion 
	    (const char *className, const char * pathName, const char *dlmessage);

	InvalidVersion 
	    (const InvalidVersion &exception);

	InvalidVersion& operator= (const InvalidVersion &);

	char *describeException (void);

       ~InvalidVersion ();
};

class NonexistentVersion: public Exception
{
	char *className;
	char *versionName;

    public:

	NonexistentVersion (const char *className, const char * versionName);
	NonexistentVersion (const NonexistentVersion &exception);
	NonexistentVersion& operator= (const NonexistentVersion &exception);
	char *describeException (void);
       ~NonexistentVersion ();
};

class NoCreateFunction: public Exception
{
	char *className;
	char *funcName;
	char *fileName;

    public:

	NoCreateFunction
	    (const char *className, const char *funcName, const char *fileName);

	NoCreateFunction (const NoCreateFunction &exception);
	NoCreateFunction& operator= (const NoCreateFunction &exception);

	char *describeException
	    (void);

       ~NoCreateFunction
	    ();
};

class NoObject: public Exception
{
    public:

	char *describeException (void);
};

#endif /* _DYN_EXCEPT_H */
