/* Copyright (c) 1992 by AT&T Bell Laboratories. */
/* Advanced C++ Programming Styles and Idioms */
/* James O. Coplien */
/* All rights reserved. */

/*
 * File bshapes.h
 * Author:  J. O. Coplien
 *
 *      ̺귯 ̴ Ŭ 
 * θ  ִ.   Լ ⿡ inline ߴ.
 */ 

#include <math.h>
#include <stdio.h>

/*-----------------------------------------------------------*/
/* Angle Ŭ    Ѵ.               */
/*  Ŭ   ϴ ý  ̽ */
/*  , е(float, double, int)  ٲ  ִ.  */
/*-----------------------------------------------------------*/

class Angle {
friend double cos(Angle);
friend double sin(Angle);
public:
    Angle(double d)       { radians = d; }
    Angle ()              { radians = 0; }
    Angle(const Angle& a) { radians = a.radians; }
    inline Angle &operator=(const Angle& a) {
        radians = a.radians;
        return *this;
    }
    Angle operator+(const Angle&) const;
private:
    double    radians;
};

/*-----------------------------------------------------------*/
/* Coordinate Ŭ īƮ 鿡  Ѵ.        */
/* ũι, ȼ, ġ, Ÿ          */
/* ǥ  ִ.                                           */ 
/*-----------------------------------------------------------*/

class Coordinate {
public:
    inline long x() const       { return x_rep; }
    inline long y() const       { return y_rep; }
    Coordinate    &operator=(const Coordinate &c) {
        x_rep = c.x_rep;
        y_rep = c.y_rep;
        return *this;
    }
    Coordinate(long x, long y) {
        x_rep = x; y_rep = y;
    }
    Coordinate(const Coordinate &c) {
        x_rep = c.x_rep;
        y_rep = c.y_rep;
    }
    Coordinate()        { x_rep = y_rep = 0;}
    ~Coordinate()       { }

    void rotateAbout(Coordinate,Angle);
private:
    long x_rep, y_rep;
};


/*-----------------------------------------------------------*/
/* Ϲ ﰢ Լ ε ؼ Angle Ű     */
/*  ֵ Ѵ.                                       */
/* ȯ ؼ ε Ұϸ,  Ű    */
/* ŸԿ ؼ ϹǷ, ﰢ Լ atan Angle Ÿ   */
/* ü ȯϵ  , ο Լ иߴ.    */
/*-----------------------------------------------------------*/

inline double cos(Angle a)      { return ::cos(a.radians); }
inline double sin(Angle a)      { return ::sin(a.radians); }
inline Angle Angleatan(double t) {
    Angle a(::atan(t)); return a;
}

/*-----------------------------------------------------------*/
/* Color Ŭ ׷ Ű ϱ  ÷ ڵ  */
/*  ĸȭѴ. δ   ̽ ƾ   */
/*   ,  ƾ  Ϲ ܰ迡    */
/* óѴ. ̷ ν پ 迡  ̽   */
/* ϰ .                                          */
/* ¦Ÿ ÷ Ӽ̴.   ϴ    */
/*  ÷  ü ̸ Ѵ.                      */
/*-----------------------------------------------------------*/

enum Flash { FlashBlack, FlashWhite, FlashOff };

class Color {
public:
    inline    double    red()         { return red_rep;  }
    inline    double    green()       { return green_rep;  }
    inline    double    blue()        { return blue_rep;  }
    Color(double r=0.0, double g=0.0,
            double b=0.0, Flash f=FlashOff) {
        red_rep = r;
        green_rep = g;
        blue_rep = b;
        flashing = f;
    }
    inline    Color&    operator=(const Color& c) {
        red_rep = c.red_rep;
        green_rep=c.green_rep;
        blue_rep = c.blue_rep;
    	return *this;
    }
    ~Color()    	{ }
private:
    double    red_rep, green_rep, blue_rep;
    Flash    flashing;
};

const Color Black, White = Color(1.0, 1.0, 1.0);

/*-----------------------------------------------------------*/
/* Shape  ׷    Ŭ̴.          */
/*  Ŭ 翡   ϸ,  翡  */
/* Ϲ ۿϴ    Ѵ(move  */ 
/*  ܸ   翡 Ϲ )             */
/*-----------------------------------------------------------*/

class Shape { 
public:
    // ߽  ȸ
    virtual    void    rotate(Angle a)    	{ } 
    virtual    void    move(Coordinate xy) {
        erase(); center_val=xy; draw();
    }
    virtual    void    draw()  = 0;     // ȭ ׸
    virtual    void    erase() {
        color = Black; draw();
    }

    virtual    void    redraw() {
        erase(); draw();
    }
    Shape();
    virtual   ~Shape();
    virtual   Coordinate origin() const  { return center_val; }
    virtual   Coordinate center() const  { return center_val; }
    inline    Shape   *next() const   { return next_pointer; }
    inline    Shape   *prev() const   { return prev_pointer; }
    inline    void    insert(Shape* i) {
        i->next_pointer= this;
        i->prev_pointer=prev_pointer;
	if (prev_pointer) {
        	prev_pointer->next_pointer=i;
	}
	prev_pointer = i;
    }
    inline    void    append(Shape* i)    {
        i->next_pointer=next();
        i->prev_pointer = this;
        if (next_pointer) {
            next_pointer->prev_pointer=i;
        }
        next_pointer = i;
    }

protected:

    /*
     *   private ſ protected ؼ
     *  Ŭ ִ ƾ   ֵ Ѵ.
     */

    Coordinate    center_val;    //  ߽
    Color    	color;    	//   ÷

private:
    Shape    	*next_pointer;   // Ʈ  
    Shape    	*prev_pointer;   // Ʈ  
};

/*-----------------------------------------------------------*/
/* Line Ư ε,   ٸ 翡  ǰ */
/* ׸µ Ѵ.  Ŭ ٸ  ȸŰµ   */
/* ⺻ ϴ rotateAbout  Լ     */
/* Ѵ. rotate  rotateAbout ȭ ε,  */
/* ߽ 糡  ϸ,  ù ̴.      */
/*-----------------------------------------------------------*/

class Line : public Shape {
public:
    void    rotate(Angle a)    { rotateAbout(center(),a); }
    void    rotateAbout(Coordinate,Angle);
    void    draw();
    inline Line &operator=(const Line &l) {
	 (void)this->Shape::operator=(*this);
         p=Coordinate(l.p); q=Coordinate(l.q); return *this;
    }
    Line(Coordinate,Coordinate);
    Line(Line& l) {
        p = l.p; q = l.q;
        center_val = l.center_val;
    }
    Line()    	    { p = Coordinate(0,0);
                      q = Coordinate(0,0); }
   ~Line()    	    { erase(); }
    inline  Coordinate      e() const   { return q; }
    inline  Coordinate origin() const   { return p; }
private:
    Coordinate    p, q;    	    //   
};

/*-----------------------------------------------------------*/
/* Rect   ȴ.  » ̸,        */
/*  ù° Ű  Ѵ. ߽  ٿ  */
/*   ̴߽.   Shape ӹ޴´.     */
/* Rep private ſ protected ؼ ڼ Ŭ     */
/* Square ̸   ֵ Ѵ.                      */
/*-----------------------------------------------------------*/

class Rect : public Shape {
public:
    void    	rotate(Angle);
    void    	draw();
    void    	move(Coordinate);
    // » , x ũ, y ũ
    Rect(Coordinate,long,long);
    ~Rect()    	    { erase(); }
    inline    Coordinate    origin() const { return l1.origin(); }
protected:
    Line    	l1, l2, l3, l4;
};

/*-----------------------------------------------------------*/
/* Ÿ ߽  ̴߽.                                */
/* Ÿ   ʾҴ.                              */
/*-----------------------------------------------------------*/

class Ellipse : public Shape {
public:
    void    	rotate(Angle);
    void    	draw();
    // ߽, , 
    Ellipse(Coordinate,long,long);
    ~Ellipse()    	{ erase(); }
protected:
    long    	major, minor;
};

/*-----------------------------------------------------------*/
/* ﰢ   Ѵ. ߽             */
/* հ Ѵ.                                      */
/*-----------------------------------------------------------*/

class Triangle : public Shape {
public:
    void    	rotate(Angle);
    void    	draw();
    void    	move(Coordinate);
    Triangle(Coordinate,Coordinate,Coordinate);
    ~Triangle()    	{ erase(); }
private:
    Line    	l1, l2, l3;
};

/*-----------------------------------------------------------*/
/* 簢 ϰ  簢̴.   Լ   */
/* Rect ӹ޴´.  ڸ Ưȭ״µ, ƹ     */
/* ۿ뵵  ʰ,  Rect ڸ ȣ ̴.   */
/*-----------------------------------------------------------*/

class Square : public Rect {
public:
    Square(Coordinate ctr, long x):
    Rect(ctr, x, x)         { }
};

