/* Copyright (c) 1992 by AT&T Bell Laboratories. */
/* Advanced C++ Programming Styles and Idioms */
/* James O. Coplien */
/* All rights reserved. */


#include "5-21.cc"

class BPF: public Filter {
public:
    Value *evaluate(Value *input = 0);
    Value *operator()(Value* f);
    BPF(Frequency w1, Frequency w2): Filter(w1, w2) { baseType = T_BPF; }
    void print() {
        if (cachedInput) cout << "COMPOUND" << endl;
        else cout << "BPF(" << f1() << ")" << endl;
    }
};

Value *
BPF::evaluate(Value *input = 0) {
    Value *f = cachedInput->evaluate(input);
    if (f->f1() > f1() && f->f1() < f2()) return f;
    else return zero;
}

Value *
BPF::operator()(Value* f) {
    // move to here to prevent gcc warning by jhpark
    Frequency lowfreq;
    Frequency highfreq;
    switch(f->type()) {
    case T_LPF:
        if (f->f1() > f2()) return this;
        else return new BPF(f1(), f->f1());
        break;
    case T_HPF:
        if (f->f1() < f1()) return this;
        else return new BPF(f->f1(), f2());
        break;
    case T_BPF:
        lowfreq = f->f1();
        highfreq = ((Filter*)f)->f2();
        if (f1() > lowfreq) lowfreq = f1();
        if (f2() < highfreq) highfreq = f2();
        return new BPF(lowfreq, highfreq);
        break;
    case T_Notch:
        cachedInput = f;
        return this;
        break;
    case T_Data:
        myType = T_Data;
        cachedInput = f;
        return evaluate();
        break;
    }
}

class Notch: public Filter {
public:
    Value *evaluate(Value *input = 0) { }
    Value *operator()(Value* f) { }
    Notch(Frequency w1, Frequency w2): Filter(w1, w2) { };
    void print() { }
};

class LPF: public Filter {
public:
    Value *evaluate(Value *input = 0);
    Value *operator()(Value* f);
    LPF(Frequency w1): Filter(w1, 0) { baseType = T_LPF; }
    void print() {
        if (cachedInput) cout << "COMPOUND" << endl;
        else cout << "LPF(" << f1() << ")" << endl;
    }
};

Value *
LPF::evaluate(Value *input = 0) {
    Value *f = cachedInput->evaluate(input);
    if (f->f1() < f1()) return f;
    else return zero;
}

Value *
LPF::operator()(Value* f) {
    // move to here to prevent gcc warning by jhpark
    switch(f->type()) {
    case T_LPF:
        if (f->f1() > f2()) return this;
        else return f;
        break;
    case T_HPF:
        if (f->f1() > f1()) return new Notch(f1(), f->f1());
        else return new BPF(f->f1(), f1());
        break;
    case T_BPF:
        if (((Filter*)f)->f2() < f1()) return f;
        else return new BPF(f->f1(), f1());
        break;
    case T_Notch:
        cachedInput = f;
        return this;
        break;
    case T_Data:
        myType = T_Data;
        cachedInput = f;
        return evaluate();
        break;
    }
}
