/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 3-6 */

#include <iostream>
#include <string>
#include <vector>

using namespace std;

// Node class definition
//
class Node {
public:
    // constructor & descructor
    Node() {
        left  = 0;
        right = 0;

        con   = new char[1];
        con[0] = '\0';
    }
   
    Node(const Node& n) {
        left = right = 0;
        if (n.left != 0)  left  = new Node(*(n.left));
        if (n.right != 0) right = new Node(*(n.right));
        
        con = new char[::strlen(n.con)+1];
        ::strcpy(con, n.con);
    }

    Node(const char* s) {
        left  = 0;
        right = 0;

        con = new char[::strlen(s) + 1];
        ::strcpy(con, s);
    }

    Node(string& s) {
        vector<string> tokens;
        tokenize(s, tokens);

        left = right = 0;
        if (tokens[0].length() > 0) left   = new Node(tokens[0]);
        if (tokens[2].length() > 0) right  = new Node(tokens[2]);
       
        con    = new char[tokens[1].length() + 1];
        ::strcpy(con, tokens[1].c_str());
    }

    ~Node() {
        if (left != 0)  delete left;
        if (right != 0) delete right;
        if (con != 0)   delete[] con;
    }

    // operator
    Node& operator=(const Node& n) {
        if (con != n.con) {
             if (left != 0) delete left;
             if (right != 0) delete right;
             if (con != 0) delete[] con;

             left = right = 0;
             if (n.left != 0)  left  = new Node(*(n.left));
             if (n.right != 0) right = new Node(*(n.right));
        
             con = new char[::strlen(n.con)+1];
             ::strcpy(con, n.con);
        }
        return *this;
    }

    // other public functions
    Node* leftChild() const { return left; }
    Node* rightChild() const { return right; }
    const char* contents() const { return con; }

private:
    // private data
    Node* left;
    Node* right;
    char* con;

    // private functions
    void   tokenize(const string&, vector<string>&); 
    string nextToken(string&);
};

void Node::tokenize(const string& s, vector<string>& tokens) {
    // ignore the first '(' and the last ')'.
    string stemp = s.substr(1, s.length() - 2);
   
    // get each token and add it to vector
    tokens.push_back(nextToken(stemp));
    tokens.push_back(nextToken(stemp));
    tokens.push_back(nextToken(stemp));
}

string Node::nextToken(string& s) {

    string retval = "";
    int    length = s.length();

    if (length == 0) {
        return retval;
    }
    else if (s[0] == ' ') {
        s = s.substr(1, length - 1);
        return retval;
    } 
    else if (s[0] == '(') {
        // skip characters till we find the matching ')'.    
        int    i, count;
        for (i = 1, count = 1; i < length && count > 0; i++) {
             if (s[i] == '(') count++;
             else if (s[i] == ')') count--;
        }
        retval = s.substr(0, i);
        s = (i == length)? "" : s.substr(i+1, length-i-1);
        return retval;
    }
    else {
        string::size_type i = s.find_first_of(" ", 1);
        retval = s.substr(0, i);
        s = (i == length)? "" : s.substr(i+1, length-i-1);
        return retval;
    }

    return retval;
}

// Functions for testing
//
string nodeWalk(const Node* n) {
    if (n == 0) return "";

    string retval = string("(") + nodeWalk(n->leftChild());
    retval = retval + " " + n->contents();
    retval = retval + " " + nodeWalk(n->rightChild()) + ")";
    return retval;	
}

void checkNode(const char* str) {
    string s(str);
    Node n(s);
    cout << "original string: " << s << endl;
    cout << "nodeWalk output: " << nodeWalk(&n) << endl << endl;
}

int main() {
   
    checkNode("(  )");
    checkNode("( a )" );
    checkNode("((  )  )");
    checkNode("(( a )  )");
    checkNode("(( a ) b )");
    checkNode("(( a ) b ( c ))");
    checkNode("( a (( b ) c ( d )))");
    checkNode("(( a ) b (( d ) e ( f )))");
    checkNode("((( a ) b ( c )) d (( e ) f ( g )))");
 
    return 0;
}
