/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 4-5 */

#define N_FLOORS      10
#define N_ELEVATORS   5

typedef int    Direction;
typedef bool   DoorStatus;
typedef int    ElevatorId;
typedef int    FloorId;

class Controller;

// Floor Control Panel class definition
//
class FloorControlPanel {
public:
    FloorControlPanel(const FloorId, const Controller*);
    ~FloorControlPanel();

    void        push(const Direction);

private:
    Controller* _controller;
    FloorId     _floor;

    bool        _up;
    bool        _down;
};

// Elevator Control Panel class definition
//
class ElevatorControlPanel {
public:
    ElevatorControlPanel(const ElevatorId, const Controller*);
    ~ElevatorControlPanel();

    void        pushFloor(const FloorId);
    void        pushEmergencyStop();
    void        pushOpenDoor();
    void        pushCloseDoor();

private:
    Controller* _controller;
    ElevatorId  _elevator;

    bool        _floor[N_FLOORS];
    bool        _emergencyStop;
    bool        _openDoor;
    bool        _closeDoor;
};

// Master Control Panel class definition
//
class MasterControlPanel {
public:
    MasterControlPanel(const Controller*);
    ~MasterControlPanel();

    void        pushEmergencyStop(const ElevatorId);
      
private:
    Controller* _controller;

    bool        _location[N_ELEVATORS][N_FLOORS];
    bool        _emergencyStop[N_ELEVATORS];
};

// Elevator class definition
//
class Elevator {
public:
   Elevator();
   ~Elevator();

    FloorId     currentFloor() const;
    DoorStatus  doorStatus() const;

    void        moveTo(const FloorId);
    void        openDoor();
    void        closeDoor();

private:
   FloorId      _currentFloor;
   DoorStatus   _doorStatus;
};

// Floor class definition
//
class Floor {
public:
    Floor();
    ~Floor();
    
    bool    upLight() const;
    bool    downLight() const;

    void    turnUpLight(const bool);
    void    turnDownLight(const bool);

private:
    bool    _upLignt;
    bool    _downLight;
};

// Controller class definition
//
class Controller {
public:
    void        request(const FloorId, const Direction);
    
    void        move(const ElevatorId, const FloorId);
    void        stop(const ElevatorId);

    void        openDoor(const ElevatorId);
    void        closeDoor(const ElevatorId);
        
private:
    Floor       _floors[N_FLOORS];
    Elevator    _elevators[N_ELEVATORS];
};
