/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 5-4 */

#include <iostream.h>
#include <string>

using namespace std;

// Atom class definition
//
class Atom {
public:
    Atom() {}
    virtual ~Atom() {}
    virtual Atom*    copy() const = 0;
    virtual operator char()  { return 0; }
    virtual long     value() { return 0; }
    virtual string   name()  { return string("error"); }
    virtual void     print() const { cout << "Atom" << endl; }
};

// NumericAtom class definition
//
class NumericAtom : public Atom {
public:
    NumericAtom() : sum(0) {}
    NumericAtom(string& s) {
        int i;
        for (i = 0, sum = 0; s[i] >= '0' && s[i] <= '9'; i++) {
            sum = (sum * 10) + s[i] - '0';
        }
        s = s.substr(i, s.length() - i);
    }
    NumericAtom(const NumericAtom& n) { sum = n.value(); }
    
    Atom* copy() const   { 
        NumericAtom *retval = new NumericAtom;
        retval->sum = sum;
        return retval;
    }

    long  value() const  { return sum; }
    void  print() const  { cout << "NumericAtom: " << sum << endl; }

private:
    long sum;
};

// Name class definition
//
class Name : public Atom {
public:
    Name() : n("")  {}
    Name(string& s) {
        int i = 0;
        for (i = 0; s[i] >= 'a' && s[i] <= 'z'; i++) {
            n = n + s.substr(i, 1);
        }
        s = s.substr(i, s.length()-i);
    }
    Name(const Name& m) { n = m.name(); }
    ~Name() {}

    Atom* copy() const {
        Name* retval = new Name;
        retval->n = n;
        return retval;
    }

    string name() const   {return n; }
    void   print() const  { cout << "Name: " << n.c_str() << endl; }

private:
    string n;
};

// Punct class definition
//
class Punct : public Atom {
public:
    Punct() : c('\0')     {}
    Punct(string& s)      { c = s[0]; s = s.substr(1,s.length()-1); }
    Punct(const Punct& p) { c = char(p); }
    ~Punct() {}
    operator char() const { return c; }

    Atom *copy() const { 
        Punct *retval = new Punct;
        retval->c = c;  
        return retval; 
    }
    void   print() const  { cout << "Punct: " << c << endl; }

private:
    char c;
};

// Oper class definition
//
class Oper : public Atom {
public:
    Oper(): c('\0') { }
    Oper(string& s) { c = s[0]; s = s.substr(1,s.length()-1); }
    Oper(Oper& o)   { c = char(o); }
    ~Oper() {}
    operator char() const { return c; }
    
    Atom *copy() const { 
        Oper *retval = new Oper;
        retval->c = c; return retval; 
    }
    void   print() const  { cout << "Oper: " << c << endl; }

private:
    char c;
};

// GeneralAtom class definition
//
class GeneralAtom : public Atom {
public:
    GeneralAtom(string&);
    GeneralAtom(const GeneralAtom& a) { realAtom = a.copy(); }
    ~GeneralAtom() { delete realAtom; }
    operator char() { return realAtom->operator char(); }  
    
    Atom* transform() {
        Atom* retval = realAtom;
        realAtom = 0;
        return retval;
    }
    Atom* copy() const {
        return realAtom->copy();
    }
    
    long   value() { return realAtom->value(); }
    string name()  { return realAtom->name(); }
   // void   print() const  { realAtom->print(); }

private:
    Atom* realAtom;
};

GeneralAtom::GeneralAtom(string& s) {
    if (!s.length()) realAtom = 0;
    else switch(s[0]) {
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            realAtom = new NumericAtom(s); 
            break;
        case ',': case ':': case '.': case ';':
            realAtom = new Punct(s);
            break;
        case '*': case '/': case '+': case '-': case '=':
            realAtom = new Oper(s); 
            break;
        default:
            if (s[0] >= 'a' && s[0] <= 'z') {
                realAtom = new Name(s);
            } else 
                realAtom = 0;
    }
}

Atom* lexAtom(string& s) {
    GeneralAtom g(s);
    return g.transform();
}

void parse(string& s) {
    for (Atom *a = lexAtom(s); a; a = lexAtom(s)) {
       a->print();
    }
}

int main() {
    string s = "x=3;y=x+5;";
    cout << "parsing original string = " << s.c_str() << endl;
    parse(s);

    return 0;
}



