/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 5-7 */

#include <iostream.h>
#include <math.h>

#define EPSILON     (1.0e-7)

// class definitions
//
class Function {
public:
    virtual double operator()(const double) = 0;
};

class Differentiator {
public:
    double operator()(Function*, const double);
};

class Integrator {
public:
   double operator()(Function*, const double, const double);
};

// class implementations
//
double Differentiator::operator()(Function* f, const double x) {
    
    // Differentiate the function using Taylor expansion
    //
    double delta, f_x, f_x_delta;
    double derivative = HUGE_VAL, oldDerivative;      
    for (delta = 1.0; true; delta *= 0.1) {
        f_x = (*f)(x);
        f_x_delta = (*f)(x + delta);
          
        oldDerivative = derivative;
        derivative = (f_x_delta - f_x) / delta;
        if (fabs(oldDerivative - derivative) < EPSILON) 
            return derivative;

        if (delta < EPSILON) 
            return derivative;
    }
}

double Integrator::operator()(Function* f, const double lower, const double upper) {
        
    // Differentiate the function using the definition
    //
    double numIntervals, delta, x, f_x;
    double sum = HUGE_VAL, oldSum;
    for (numIntervals = 1.0; true; numIntervals *= 10) {
        delta = (upper - lower) / numIntervals;

        oldSum = sum; sum = 0;
        for (x = lower; x <= upper; x += delta) {
            f_x = (*f)(x);
            sum += f_x * delta;
        }
        if (fabs(oldSum - sum) <= EPSILON) 
            return sum;
        if (delta < EPSILON) 
            return sum;
    }
}

// sample classes
//
class F : public Function {
public:
    double operator()(const double x) { return x*x; }
};

class G : public Function {
public:
    double operator()(const double x) { return (x*x + 2*x + 1); }
};

int main() {

    F               f;
    G               g;
    Differentiator  D;
    Integrator      I;

    cout << "f(1) = " << f(1) << endl;
    cout << "f'(1) = " << D(&f, 1) << endl;
    cout << "integrate f from (0, 1)= " << I(&f, 0, 1) << endl;

    cout << "g(1) = " << g(1) << endl;
    cout << "g'(1) = " << D(&g, 1) << endl;
    cout << "integrate f from (0, 1)= " << I(&g, 0, 1) << endl;

    return 0;
}
