/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 6-3 */

#include <iostream.h>
#include <string.h>

#define MAX(x, y)   (((x) > (y))? x: y)

class Account {
public:
    Account() { 
        penalty = minimum = balance = 0;
        name    = new char[8];
        strcpy(name, "Account");
    }
    Account(double penalty, double minimum, 
            const char* s="Account") {
        this->penalty = penalty;
        this->minimum = minimum;
        this->balance = 0;
        this->name    = new char[strlen(s) + 1];
        strcpy(this->name, s);

    }
        
    virtual ~Account() { 
        if (name) delete[] name; 
    }

    virtual void deposit(double) = 0;
    virtual void withdraw(double) = 0;
    virtual void print_statement() = 0;
    
protected:
    void _deposit(double amount)   { 
        cout << "deposit " << amount << " to " << name << endl;
        balance += amount; 
    }

    void _withdraw(double amount)  { 
        if (balance < amount) 
            cout << "balance is not enough. can not withdraw." << endl;
        else {
            cout << "withdraw " << amount << " from " << name << endl;
            balance -= amount; 
        }

        if (balance < minimum) { 
            balance = MAX(balance - penalty, 0);
            cout << "lower than minimum balance. penalty is applied. " << endl;
        }
    }
    void _print_statement() { 
        cout << "[Account Statement]" << endl;
        cout << "NAME        " << name << endl;
        cout << "BALANCE     " << balance << endl;
        cout << "MINIMUM     " << minimum << endl;
        cout << "PENALTY     " << penalty << endl;
    }

private:
    double penalty;
    double minimum;
    double balance;
    char*  name;
};

class InterestAccount : virtual public Account {
public:
    void deposit(double amount){ 
        Account::_deposit(amount);
        _deposit(amount);
    }
    void withdraw(double amount){
        Account::_withdraw(amount);
        _withdraw(amount);
    }
    void print_statement(){
        Account::_print_statement();
        _print_statement();
    }
    double calculate_interest(){
        // currently, there is no interest.
        return 0.0L;
    }

protected:
    InterestAccount(double penalty, double interest, double minimum, 
                    const char* s = "InterestAccount") :
                    Account(penalty, minimum, s) { 
      
        this->interest = interest;
    }

    virtual ~InterestAccount() {}

    void _deposit(double amount){ }
    void _withdraw(double amount){ }
    void _print_statement(){
        cout << "INTEREST    " << interest << endl;
    }

private:
    double interest;
};

class SavingsAccount : public InterestAccount {
public:
    SavingsAccount(double penalty, double interest, double minimum,
                   const char* s = "SavingsAccount"):
                   InterestAccount(penalty, interest, minimum, s) {}
    
    virtual ~SavingsAccount(){}
private:
};

class CheckingAccount : virtual public  Account {
public:
    CheckingAccount(double penalty, double minimum, 
        const char* s = "CheckingAccount") : 
        Account(penalty, minimum, s) { }
       
    virtual ~CheckingAccount() {};

    void deposit(double amount){ 
        Account::_deposit(amount);
        _deposit(amount);
    }
    void withdraw(double amount){
        Account::_withdraw(amount);
        _withdraw(amount);
    }
    void print_statement(){
        Account::_print_statement();
        _print_statement();
    }

protected:
    void _deposit(double){}
    void _withdraw(double){}
    void _print_statement(){}

private:
};


class InterestCheckingAccount : public CheckingAccount, public InterestAccount {
public:
    InterestCheckingAccount(double penalty, double interest, double minimum,
        const char* s = "InterestCheckingAccount") :
        Account(penalty, minimum, s),
        CheckingAccount(penalty, minimum, s),
        InterestAccount(penalty, interest, minimum, s) {
    }
    
    void deposit(double amount){
        Account::_deposit(amount);
        CheckingAccount::_deposit(amount);
        InterestAccount::_deposit(amount);
        _deposit(amount);
    }
    void withdraw(double amount){
        Account::_withdraw(amount);
        CheckingAccount::_withdraw(amount);
        InterestAccount::_withdraw(amount);
        _withdraw(amount);
    }
    void print_statement(){
        Account::_print_statement();
        CheckingAccount::_print_statement();
        InterestAccount::_print_statement();
        _print_statement();
    }

protected:
    void _deposit(double){}
    void _withdraw(double){}
    void _print_statement(){}

private:
};                          

int main() {
    InterestCheckingAccount acct(10, 0.1, 1000, "sample-interest-checking-account");
    acct.print_statement();
    acct.deposit(2000);
    acct.print_statement();
    acct.withdraw(1200);
    acct.print_statement();

    return 0;
}
