/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 8-1 */

#include <iostream.h>
#include <string>

using namespace std;

class Exemplar { public: Exemplar() { } };

// Atom class definition
//
class Atom {
protected: 
    static  Atom*   list;
    Atom*   next;
    static  char    errFlag;

public:
    Atom()         { next = 0; errFlag = 0; }
    Atom(Exemplar) { next = list; list = this; }
    virtual ~Atom(){}
    virtual Atom* make(string& s) {
        Atom* retval = 0;
        extern Atom* atomExemplar;
        for (Atom* a = list; a; a = a->next) {
            if (a != atomExemplar)
                if (retval = a->make(s)) break;           
        }
        return retval;
    }

    virtual operator char()  { return 0; }
    virtual long     value() { return 0; }
    virtual string   name()  { return string("error"); }
    virtual void     print() const { cout << "Atom" << endl; }
};

Atom* Atom::list    = 0;
char  Atom::errFlag = 0;

Atom *atomExemplar = new Atom(Exemplar());

// Number class definition
//
class Number : public Atom {
public:
    Number(Exemplar a) : Atom(a), sum(0) {}
    Number(string& s) : Atom(), sum(0) {
        Atom::errFlag = (s[0] < '0' || '9' < s[0]);
        if (Atom::errFlag) return;

        int i;
        for (i = 0, sum = 0; s[i] >= '0' && s[i] <= '9'; i++) {
            sum = (sum * 10) + s[i] - '0';
        }
        s = s.substr(i, s.length() - i);
    }
    Number(const Number& n) { sum = n.value(); }
    ~Number() {}

    Atom* make(string& s) { 
        Atom *retval = new Number(s);
        if (Atom::errFlag) { delete retval; retval = 0; }
        return retval;
    }

    long  value() const  { return sum; }
    void  print() const  { cout << "Number: " << sum << endl; }

private:
    long sum;
};

static Number *numberInstance = new Number(Exemplar());

// Name class definition
//
class Name : public Atom {
public:
    Name(Exemplar a) : Atom(a), n("")  {}
    Name(string& s) : Atom(), n("") {
        errFlag = (s[0] < 'a' || 'z' < s[0]);
        if (errFlag) return;

        int i;
        for (i = 0; s[i] >= 'a' && s[i] <= 'z'; i++) {
            n = n + s.substr(i, 1);
        }    
        s = s.substr(i, s.length()-i);
    }
    Name(const Name& m) { n = m.name(); }
    ~Name() {}

    Atom* make(string& s) {
        Atom* retval = new Name(s);
        if (Atom::errFlag) { delete retval; retval = 0; }
        return retval;
    }

    string name() const   { return n; }
    void   print() const  { cout << "Name: " << n.c_str() << endl; }

private:
    string n;
};

static Name *nameInstance = new Name(Exemplar());

// Punct class definition
//
class Punct : public Atom {
public:
    Punct(Exemplar a) : Atom(a), c('\0')   {}
    Punct(string& s) : Atom(), c('\0')     { 
        errFlag = !ispunct(s[0]);
        if (errFlag) return;

        c = s[0]; 
        s = s.substr(1,s.length()-1); 
    }
    Punct(const Punct& p) { c = char(p); }
    ~Punct() {}

    Atom *make(string& s) { 
        Atom *retval = new Punct(s);
        if (errFlag) { delete retval; retval = 0; }
        return retval; 
    }
   
   operator char() const { return c; }
   void   print() const  { cout << "Punct: " << c << endl; }

private:
    char c;
};

static Punct *punctInstance = new Punct(Exemplar());

// Oper class definition
//
class Oper : public Atom {
public:
    Oper(Exemplar a) : Atom(a), c('\0') { }
    Oper(string& s) : Atom(), c('\0')  { 
        errFlag = (s[0] != '+' && s[0] != '-' && s[0] != '*' && s[0] != '/'&& s[0] != '=');
        if (errFlag) return;

        c = s[0]; 
        s = s.substr(1,s.length()-1); 
    }
    Oper(Oper& o)   { c = char(o); }
    ~Oper() {}

    Atom *make(string& s) { 
        Atom *retval = new Oper(s);
        if (errFlag) { delete retval; retval = 0; }
        return retval; 
    }
    operator char() const { return c; }
    void   print() const  { cout << "Oper: " << c << endl; }

private:
    char c;
};

static Oper *operInstance = new Oper(Exemplar());

// Functions for testing
//
void parse(string& s) {
    while (s.length() > 0) {
        Atom* a = atomExemplar->make(s);
        if (a) { a->print(); delete a; }
    }
}

int main() {
    string s= "x=3;y=x+5";
    cout << "parsing original string:" << s.c_str() << endl;
    parse(s);

    return 0;
}
