/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 8-2 */

#include <iostream.h>
#include <string>
#include <vector>
#include <string.h>

using namespace std;

// Exemplar name server
//
class Atom;
template<class T> class NameServer {
public:
    NameServer() { _iter = 0; }
    T* lookup(const char* name) {
        if (_iter == 0 || _iter == _exemplars.end()) return lookupNew(name);
        if (strcmp(name, (*_iter)->sname()))         return lookupNew(name);
        return (*_iter);
    }
    void  registerExemplar(T* exemplar) { 
        _exemplars.push_back(exemplar);
        cout << "registering exemplar(" << exemplar->sname() << ")..." << endl;
    }

private:
    vector<T*>            _exemplars;
    vector<T*>::iterator  _iter;

    T* lookupNew(const char* name) {
        cout << "iterate over all exemplars to find '" << name << "'" << endl;
        for (_iter = _exemplars.begin(); _iter != _exemplars.end(); _iter++)
            if (!strcmp(name, (*_iter)->sname())) return (*_iter);
        return 0;
    }
};

NameServer<Atom>* atomNameServer = new NameServer<Atom>;

// Sample parser abstract base exemplar application
//
class Exemplar { public: Exemplar() { } };

class Atom {
protected:
    static char errFlag;
public:
    virtual ~Atom() {}
    Atom() {}
    Atom(Exemplar) {}
    virtual Atom* make(string& s) {
        Atom* a = 0;
        while (true) {
            if (a = atomNameServer->lookup("name")->make(s)) break;
            if (a = atomNameServer->lookup("number")->make(s)) break;
            if (a = atomNameServer->lookup("punct")->make(s)) break;
            if (a = atomNameServer->lookup("oper")->make(s)) break;
        }
        return a;
    }
    virtual operator char()  { return '\0'; }
    virtual long     value() { return 0; }
    virtual string   name()  { return "atom"; }
    virtual void     print() const { cout << "error" << endl; };
    virtual char*    sname() const { return "atom"; }
};

char Atom::errFlag = 0;

Atom *atomExemplar = new Atom(Exemplar());

class Number : public Atom {
public:
    Number(Exemplar a) : sum(0) { atomNameServer->registerExemplar(this); }
    Number(string& s) : sum(0)  {
        Atom::errFlag = (s[0] < '0' || '9' < s[0]);
        if (Atom::errFlag) return;

        int i;
        for (i = 0, sum = 0; s[i] >= '0' && s[i] <= '9'; i++) {
            sum = (sum * 10) + s[i] - '0';
        }
        s = s.substr(i, s.length() - i);
    }
    Number(const Number& n) { sum = n.value(); }
    ~Number() {}

    Atom* make(string& s) { 
        Atom *retval = new Number(s);
        if (Atom::errFlag) { delete retval; retval = 0; }
        return retval;
    }

    long  value() const  { return sum; }
    void  print() const  { cout << "Number: " << sum << endl; }
    char* sname() const  { return "number"; }

private:
    long sum;
};

static Number *numberInstance = new Number(Exemplar());

class Name : public Atom {
public:
    Name(Exemplar a) : n("")  { atomNameServer->registerExemplar(this); }
    Name(string& s) : n("")   {
        errFlag = (s[0] < 'a' || 'z' < s[0]);
        if (errFlag) return;

        int i;
        for (i = 0; s[i] >= 'a' && s[i] <= 'z'; i++) {
            n = n + s.substr(i, 1);
        }    
        s = s.substr(i, s.length()-i);
    }
    Name(const Name& m) { n = m.name(); }
    ~Name() {}

    Atom* make(string& s) {
        Atom* retval = new Name(s);
        if (Atom::errFlag) { delete retval; retval = 0; }
        return retval;
    }

    string name() const   { return n; }
    void   print() const  { cout << "Name: " << n.c_str() << endl; }
    char*  sname() const  { return "name"; }

private:
    string n;
};

static Name *nameInstance = new Name(Exemplar());

class Punct : public Atom {
public:
    Punct(Exemplar a) : c('\0')    { atomNameServer->registerExemplar(this); }
    Punct(string& s) : c('\0')     { 
        errFlag = !ispunct(s[0]);
        if (errFlag) return;

        c = s[0]; 
        s = s.substr(1,s.length()-1); 
    }
    Punct(const Punct& p) { c = char(p); }
    ~Punct() {}

    Atom *make(string& s) { 
        Atom *retval = new Punct(s);
        if (errFlag) { delete retval; retval = 0; }
        return retval; 
    }
   
   operator char() const { return c; }
   void   print() const  { cout << "Punct: " << c << endl; }
   char*  sname() const  { return "punct"; }

private:
    char c;
};

static Punct *punctInstance = new Punct (Exemplar());

class Oper : public Atom {
public:
    Oper(Exemplar a) : c('\0') { atomNameServer->registerExemplar(this); }
    Oper(string& s) : c('\0')  { 
        errFlag = (s[0] != '+' && s[0] != '-' && s[0] != '*' && s[0] != '/'&& s[0] != '=');
        if (errFlag) return;

        c = s[0]; 
        s = s.substr(1,s.length()-1); 
    }
    Oper(Oper& o)   { c = char(o); }
    ~Oper() {}

    Atom *make(string& s) { 
        Atom *retval = new Oper(s);
        if (errFlag) { delete retval; retval = 0; }
        return retval; 
    }
    operator char() const { return c; }
    void   print() const  { cout << "Oper: " << c << endl; }
    char*  sname() const  { return "oper"; }

private:
    char c;
};

static Oper operInstance(Exemplar());

// Functions for testing
//
void parse(string& s) {
    while (s.length() > 0) {
        Atom* a = atomExemplar->make(s);
        if (a) { a->print(); delete a; }
    }
}

// Functions for testing
//
void makeAtom(const char* name, string& s) {
    Atom* a = atomNameServer->lookup(name)->make(s);
    a->print(); delete a;
}

int main() {
    string s= "x=3;y=x+5";
    cout << endl << "parsing original string:" << s.c_str() << endl;
    parse(s);

    return 0;
}
