/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 8-4 */

#include <iostream.h>
#include <string>

using namespace std;

class Exemplar { public: Exemplar() {} };

// Class class definition
class Class {
public:
    enum Directive { PrintPayCheck, GiveRaise, 
                     Name, TimeWorked, Salary, StockOptions, 
                     ToString, ToInt, Value };

    virtual Class* atPut(Directive, Class*) = 0;
    virtual Class* at(Directive) = 0;
    virtual Class* getNext() = 0;
    virtual Class* putNext(Class*)  = 0;
    virtual Class* make(Class*) = 0; 
    virtual Class* make(int) { return 0; }
    virtual Class* make(string&, int=0, int=0, int=0)  { return 0; } 
};

// Integer class definition
extern class Class* integerExemplar;

class Integer : public Class {
public:
    Integer(Exemplar)     { }

    Class* atPut(Directive key, Class* arg) {
        switch(key) {
        case Value: 
            _rep = (int)arg->at(ToInt); 
            return this;
        default:
            return this;
        }
    }

    Class* at(Directive key) {
        switch(key) {
        case Value: 
            return integerExemplar->make(_rep);
        case ToInt:
            return (Class*)_rep;
        default:
            return this;
        }
    }

    Class* getNext() { return 0; }
    Class* putNext(Class*) { return 0; }
    Class* make(int i) {
        Class* n = new Integer(i);
        return n;
    }

    Class* make(Class* param) {
        Class*  n = new Integer((int)param->at(ToInt));
        return n;
    }

protected:
    Integer(const int i)  { _rep = i; }

private:
    int _rep;
};

Class* integerExemplar = new Integer(Exemplar());

// String class definition
extern class Class* stringExemplar;

class String : public Class {
public:
    String(Exemplar) {}
    
    Class* atPut(Directive key, Class* arg) {
        switch(key) {
        case Value: 
            _rep = *(string*)arg->at(ToString);
            return this;
        default:
            return this;
        }
    }

    Class* at(Directive key) {
        switch(key) {
        case Value: 
            return stringExemplar->make(_rep);
        case ToString:
            return (Class*)&_rep;
        default:
            return this;
        }
    }
    
    Class* getNext() { return 0; }
    Class* putNext(Class*) { return 0; }
    
    Class* make(string&s, int i=0, int j=0, int k=0) {
        Class* c = new String(s);
        return c;
    }
    Class* make(Class* param) {
        Class*  s = new String(*(string *)param->at(ToString));
        return s;
    }
    
private:
    string _rep;
    String(const string& s)  { _rep = s; }
};

Class* stringExemplar = (Class*)new String(Exemplar());

// Employee class definition
class Employee: public Class {  
public:
    virtual Class* atPut(Directive key, Class* arg) {
        switch(key) {
        case Name:
            _name = arg->at(Value);
            return this;
        case Salary:
            _salary = arg->at(Value);
            return this;
        case TimeWorked:
            _hours = arg->at(Value);
            return this;        
        default:
            return this;
        }   
    }

    virtual Class* at(Directive key) {
        switch(key) {
        case Name:
            return integerExemplar->make(_name);
        case Salary: 
            return integerExemplar->make(_salary);
        case TimeWorked:
            return integerExemplar->make(_hours);
        case PrintPayCheck: {
            cout << "[pay check]" << endl;
            string* s = (string *)(_name->at(ToString));
            cout << "name   : " << s->c_str() << endl;
            cout << "salary : " << (int)(_salary->at(ToInt)) << endl;
            cout << "hours  : " << (int)(_hours->at(ToInt)) << endl; 
            return this; }
        default:
            return this;
        }
    }

    virtual Class* getNext() = 0;
    virtual Class* putNext(Class*) = 0;
    virtual Class* make(string&, int=0, int=0, int=0) = 0;
    virtual Class* make(Class*) = 0;

protected:
    int       _pos;
    Class*    _name;
    Class*    _salary;
    Class*    _hours;

    Employee() { _pos = 0; }
    Employee(string& name, int salary, int hours) {
        _pos    = -1;
        _name   = stringExemplar->make(name);
        _salary = integerExemplar->make(salary);
        _hours  = integerExemplar->make(hours);
    }

private:
};

// VicePresident class definition
//
extern class Class* vpExemplar;

class VicePresident : public Employee {
public:
    VicePresident(Exemplar) { }
   
    Class* atPut(Directive key, Class* arg) {      
        switch(key) {
        case StockOptions:
            _stockOptions = arg->at(Value);
            return this;
        case GiveRaise:  {
            int salaryTotal = (int)arg->at(ToInt) + (int)_salary->at(ToInt);
            Employee::atPut(Salary, integerExemplar->make(salaryTotal));
            return this; }
        default:
            return Employee::atPut(key, arg);
        }
    }
    Class* at(Directive key) {
        switch(key) {
        case StockOptions:
            return integerExemplar->make(_stockOptions);
        case PrintPayCheck:
            Employee::at(key);
            cout << "stock optoins: " << (int)_stockOptions->at(ToInt) << endl;
            return this;
        default:
            return Employee::at(key);
        }
    }

    Class* getNext() { 
        ++_pos; _pos %= 4;
        if (_pos == 0) return _name;
        if (_pos == 1) return _salary;
        if (_pos == 2) return _hours;
        if (_pos == 3) return _stockOptions;
        return 0;
    }
    Class* putNext(Class* param) { 
        ++_pos; _pos %= 4;
        if (_pos == 0) return atPut(Name, param);
        if (_pos == 1) return atPut(Salary, param);
        if (_pos == 2) return atPut(TimeWorked, param);
        if (_pos == 3) return atPut(StockOptions, param);
        return 0;
    }

    Class* make(Class* param) {
        string* name         = (string*)param->getNext()->at(ToString);
        int     salary       = (int)param->getNext()->at(ToInt);
        int     hours        = (int)param->getNext()->at(ToInt);
        int     stockOptions = (int)param->getNext()->at(ToInt);

        return make(*name, salary, hours, stockOptions);
    }

    Class* make(string& s, int i, int j, int k) {
        Class* v = new VicePresident(s, i, j, k);
        return v;
    }
private:
    Class*  _stockOptions;

    VicePresident(string& name, int salary, int hours, int stockOptions) :
        Employee(name, salary, hours) {
        _stockOptions  = integerExemplar->make(stockOptions); 
    }

};
Class* vpExemplar = (Class *)new VicePresident(Exemplar());

// Function for testing
//
int main() {
    string name1     = "John Doe";
    Class* employee1 = vpExemplar->make(name1, 1000, 40, 100000);
    employee1->at(Class::PrintPayCheck);
    
    Class* employee2 = vpExemplar->make(employee1);
    Class* salaryRaise = integerExemplar->make(1000);
    employee2->atPut(Class::GiveRaise, salaryRaise);
    employee2->at(Class::PrintPayCheck);

    return 0;
}
