/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise 9-10 */

#include <iostream.h>
#include <new>

// Top & Thing class definition
//
class Top {
public:
    virtual ~Top() {}
    virtual Top* type() { 
        return this; 
    }
    static void operator delete(void *p) { ::operator delete(p); }

protected:
    Top() {}
    static void* operator new(size_t l) { return ::operator new(l); }
};

typedef unsigned long REF_TYPE;
class Thing : public Top {
public:
    Thing(): refCountVal(1), updateCountVal(0) {}
    virtual ~Thing() {}
    
    virtual REF_TYPE deref() { return --refCountVal; }
    virtual REF_TYPE ref()   { return ++refCountVal; }
    virtual Thing* cutover(Thing*) { return this; }
    
private:
    REF_TYPE refCountVal, updateCountVal;
};

// NumberRep class & its derived classes definition
//
class Number;
class NumberRep : public Thing {
public:
    NumberRep()  {}
    virtual ~NumberRep() { }
    virtual Number  make(int);
    virtual Number  make(double);
    virtual Number  make(double, double);

    virtual Thing* type() const;
    virtual bool   isA(const Number&) const;
    virtual Number promote(const Number&) const;
    virtual Number add(const Number&) const;

    virtual double rvalue() const { return 0.0; }
    virtual double ivalue() const { return 0.0; }
    virtual void   print() {}

protected:
    friend class Number;
    static void* operator new(size_t l)  { return ::operator new(l); }
    static void  operator delete(void* p) { ::operator delete(p); }    
};

static NumberRep numberRepExemplar;
NumberRep* numberRep = &numberRepExemplar;

class Integer : public NumberRep {
public:
    Integer()  { rep = 0; }
    ~Integer() { }
    Number make(int);

    Thing* type() const;
    bool   isA(const Number&) const;
    Number promote(const Number&) const;
    Number add(const Number&) const;

    double rvalue() const { return rep; }
    void   print() { cout << "integer(" << rep << ")" << endl; }

protected:
    Integer(int i) { rep = i; }

private:
    int rep;
};

static Integer  integerExemplar;
Integer* integer = &integerExemplar;

class Real : public NumberRep {
public:
    Real()  { rep = 0.0; }
    ~Real() { }
    Number make(double); 

    Thing* type() const;
    bool   isA(const Number&) const;
    Number promote(const Number&) const;
    Number add(const Number&) const;

    double rvalue() const { return rep; }
    void   print() { cout << "real(" << rep << ")" << endl; }

protected:
    Real(double r) { rep = r; }

private:
    double rep;
};

static Real  realExemplar;
Real* real = &realExemplar;

class Complex : public NumberRep {
public:
    Complex()  { rep_r = rep_i = 0.0; }
    ~Complex() { }
    Number make(double, double);

    Thing* type() const;
    bool   isA(const Number&) const;
    Number promote(const Number&) const;
    Number add(const Number&) const;

    double rvalue() const { return rep_r; }
    double ivalue() const { return rep_i; }
    void   print() { cout << "complex(" << rep_r << " + " << rep_i << " i)" << endl; }

protected:
    Complex(double r, double i) { rep_r = r; rep_i = i; }

private:
    double  rep_r, rep_i;
};

static Complex  complexExemplar;
Complex* complex = &complexExemplar;

// Number class definition
//
class Number : public Top {
public:
    Number()                { rep = (numberRep->make(0)).rep; }
    Number(const Number& n) { (rep = n.rep)->ref(); }
    Number(NumberRep& n)    { (rep = &n)->ref(); }
    ~Number()               { if (rep && rep->deref() <=0) delete rep; }

    NumberRep* operator->() const { return rep; }
    Number& operator=(const Number& n) { 
        if (rep == n.rep) return *this;
        if (rep && rep->deref() <= 0) delete rep;
        (rep = n.rep)->ref();
        return *this;
    }
    friend Number operator+(const Number&, const Number&);
private:
    static void* operator new(size_t) { cerr << "heap number" << endl; }
    static void  operator delete(void*) {}
    NumberRep*   rep;
};

static Number numberExemplar;
Number* number = &numberExemplar;

// NumberRep & its derived classes implementation
//
Number NumberRep::make(int i) {
    return integer->make(i);
}
Number NumberRep::make(double r) {
    return real->make(r);
}
Number NumberRep::make(double r, double i) {
    return complex->make(r, i);
}
Thing* NumberRep::type() const { 
    return numberRep; 
}
bool NumberRep::isA(const Number& n) const { 
    return (n->type() == integer || n->type() == real || n->type() == complex); 
}
Number NumberRep::promote(const Number& n) const {
    return Number(n); // return dummy
}

Number NumberRep::add(const Number& n) const {
    return Number(n); // return dummy
}

Number Integer::make(int i) {
    NumberRep* rep = new Integer(i);
    return Number(*rep);
}
Thing* Integer::type() const { 
    return integer; 
}
bool Integer::isA(const Number& n) const { 
    return (n->type() == integer); 
}
Number Integer::promote(const Number& n) const {
    return Number(n);
}
Number Integer::add(const Number& n) const {
    int sum = rep + (int)n->rvalue();
    return (numberRep->make(sum));
}

Number Real::make(double r) {
    NumberRep* rep = new Real(r);
    return Number(*rep);
}
Thing* Real::type() const { 
    return real;
}
bool Real::isA(const Number& n) const { 
    return (n->type() == integer || n->type() == real ); 
}
Number Real::promote(const Number& n) const {
    if (n->type() == integer) 
        return numberRep->make(n->rvalue());
    return Number(n);
}
Number Real::add(const Number& n) const {
    double sum = rep + n->rvalue();
    return (numberRep->make(sum));
}

Number Complex::make(double r, double i) {
    NumberRep* rep = new Complex(r, i);
    return Number(*rep);
}
Thing* Complex::type() const { return complex; }
bool Complex::isA(const Number& n) const {
    return (n->type() == integer || n->type() == real || n->type() == complex); 
}

Number Complex::promote(const Number& n) const {
    if (n->type() == integer || n->type() == real)
        return numberRep->make(n->rvalue(), 0.0);
    return Number(n);
}

Number Complex::add(const Number& n) const {
    double rsum = rep_r + n->rvalue();
    double isum = rep_i + n->ivalue();
    return (numberRep->make(rsum, isum));
}

// Global operator+ implementation
//
Number operator+(const Number& n1, const Number& n2) {
    if (n1->isA(n2)) {
        Number temp = n1->promote(n2);
        return n1->add(temp);
    }
    else if (n2->isA(n1)) {
        Number temp = n2->promote(n1);
        return n2->add(temp);
    }
    else
       // throw number type error
       throw;  
}

// Function for testing
//
int main() {

    Number  i = (*number)->make(1);   
    cout << "i = "; i->print();

    Number r = (*number)->make(0.5);
    cout << "r = "; r->print();

    Number c = (*number)->make(1.5, 2.3);
    cout << "c = "; c->print();

    Number x = i + r;
    cout << "i + r = "; x->print();

    Number y = c + i;
    cout << "c + i = "; y->print();

    return 0;
}

