/* Copyright (c) 2003 by Hanbit Media Co. All rights reserved*/
/* Lee, Hae-Young(hae_lee@hanmail.net) */
/* Advanced C++ Programming Styles and Idioms, exercise appendix A-1 */

#include <iostream.h>
#include <string.h>

#define POINTER(start, index, size)  ((void *)((int)start + size * index))

void exchange(void* a, void* b, int elementSize) {
        char* temp  = new char[elementSize];
        memcpy(temp, a, elementSize);
        memcpy(a, b, elementSize);
        memcpy(b, temp, elementSize);
        delete temp;
}

int partition(void* base, int nel, 
              int elementSize, int (*compar)(void*, void*)) {
    
    char* pivot = new char[elementSize];
    memcpy(pivot, base, elementSize);
   
    int i = 0, j = nel - 1;
    while (true) {
        while (compar(POINTER(base, j, elementSize), (void*)pivot) < 0) j--;
        while (compar(POINTER(base, i, elementSize), (void*)pivot) > 0) i++;
        
        if (i >= j) break;
        exchange(POINTER(base, i, elementSize), POINTER(base, j, elementSize), elementSize);
        i++; j--;
    }
    delete pivot;
    return j;
}

void qsort(void *base, int nel, 
           int elementSize, int (*compar)(void*, void*))  {
    if (nel <= 1) return;

    int index = partition(base, nel, elementSize, compar) + 1;
    qsort(base, index, elementSize, compar);
    qsort(POINTER(base, index, elementSize), nel - index, elementSize, compar);    
}

// function for testing
//
int compareInt(void* a, void* b) { 
    if (*(int*)a < *(int*)b) return 1;
    else if (*(int*)a > *(int*)b) return -1;
    else return 0;
}

void printArrayInt(int* a, int nel, const char* s=0) {
    if (s) cout << s;
    for (int i = 0; i < nel; i++) cout << a[i] << " ";
    cout << endl;
}

int main() {
    int a[10] = { 4, 1, 5, 7, 3, 10, 2, 1, 9, 6 };
    printArrayInt(a, 8, "original array: ");
    qsort(a, 8, sizeof(int), compareInt);
    printArrayInt(a, 8, "sorted ascending: ");
       
    return 0;
}
